import math
import os
import os.path as osp
import random

import numpy as np
import open3d as o3d
import cv2
import tempfile


def image_bytes_to_point_cloud(
    rgb_bytes: bytes, depth_bytes: bytes, fov: float, local_to_world_matrix: np.ndarray
):
    """
    Use the raw bytes of RGB image and depth image, as well as the camera
    FOV and extrinsic matrix to generate point cloud in global coordinate.

    Args:
        rgb_bytes: Bytes, raw bytes of RGB image.
        depth_bytes: Bytes, raw bytes of depth image.
        fov: Float, camera Field Of View (FOV).
        local_to_world_matrix: Numpy.ndarray, the local_to_world_matrix of camera.

    Return:
        open3d.geometry.PointCloud: The point cloud.
    """
    image_rgb = np.frombuffer(rgb_bytes, dtype=np.uint8)
    image_rgb = cv2.imdecode(image_rgb, cv2.IMREAD_COLOR)
    image_rgb = cv2.cvtColor(image_rgb, cv2.COLOR_BGR2RGB)

    temp_file_path = osp.join(
        tempfile.gettempdir(), f"temp_img_{int(random.uniform(10000000, 99999999))}.exr"
    )
    with open(temp_file_path, "wb") as f:
        f.write(depth_bytes)
    image_depth = cv2.imread(temp_file_path, cv2.IMREAD_UNCHANGED)
    os.remove(temp_file_path)

    pcd = image_array_to_point_cloud(image_rgb, image_depth, fov, local_to_world_matrix)
    return pcd


def image_array_to_point_cloud(
    image_rgb: np.ndarray,
    image_depth: np.ndarray,
    fov: float,
    local_to_world_matrix: np.ndarray,
):
    """
    Use the RGB image and depth image, as well as the camera
    FOV and extrinsic matrix to generate point cloud in global coordinate.

    Args:
        image_rgb: Numpy.ndarray, in shape (H,W,3), the RGB image.
        image_depth: Numpy.ndarray, in shape (H,W,3), the depth image.
        fov: Float, camera Field Of View (FOV).
        local_to_world_matrix: Numpy.ndarray, the local_to_world_matrix of camera.

    Return:
        open3d.geometry.PointCloud: The point cloud.
    """
    points = depth_to_point_cloud(image_depth, fov=fov, organized=False)
    pcd = o3d.geometry.PointCloud()
    pcd.points = o3d.utility.Vector3dVector(points)

    colors = image_rgb.reshape(-1, 3)
    pcd.colors = o3d.utility.Vector3dVector(colors.astype(np.float32) / 255.0)

    pcd.transform(local_to_world_matrix)

    return pcd


def depth_to_point_cloud(depth: np.ndarray, fov: float, organized=False):
    """
    Use the depth image and the camera FOV to generate point cloud in
    camera coordinate.

    Args:
        depth: Numpy.ndarray, in shape (H,W,3), the depth image.
        fov: Float, camera Field Of View (FOV).
        organized: Bool, whether keep the point organized. If True, the returned shape of point cloud is (H,W,3); If False, the returned shape of point cloud is (H*W,3).

    Return:
        open3d.geometry.PointCloud: The point cloud.
    """
    height = depth.shape[0]
    width = depth.shape[1]

    cy = height / 2
    cx = width / 2

    xmap = np.arange(width)
    ymap = np.arange(height)
    xmap, ymap = np.meshgrid(xmap, ymap)

    fx = fy = height / (2 * math.tan(math.radians(fov / 2)))

    points_z = depth
    points_x = (xmap - cx) * points_z / fx
    points_y = (ymap - cy) * points_z / fy

    cloud = np.stack([points_x, -points_y, points_z], axis=-1)
    if not organized:
        cloud = cloud.reshape([-1, 3])

    return cloud


def image_bytes_to_point_cloud_intrinsic_matrix(
    rgb_bytes: bytes,
    depth_bytes: bytes,
    intrinsic_matrix: np.ndarray,
    local_to_world_matrix: np.ndarray,
):
    """
    Use the raw bytes of RGB image and depth image, as well as the camera
    intrinsic matrix and extrinsic matrix to generate point cloud in global coordinate.

    Args:
        rgb_bytes: Bytes, raw bytes of RGB image.
        depth_bytes: Bytes, raw bytes of depth image.
        intrinsic_matrix: Numpy.ndarray, the intrinsic matrix of camera.
        local_to_world_matrix: Numpy.ndarray, the local_to_world_matrix of camera.

    Return:
        open3d.geometry.PointCloud: The point cloud.
    """
    image_rgb = np.frombuffer(rgb_bytes, dtype=np.uint8)
    image_rgb = cv2.imdecode(image_rgb, cv2.IMREAD_COLOR)
    image_rgb = cv2.cvtColor(image_rgb, cv2.COLOR_BGR2RGB)
    color = o3d.geometry.Image(image_rgb)

    temp_file_path = osp.join(
        tempfile.gettempdir(), f"temp_img_{int(random.uniform(10000000,99999999))}.exr"
    )
    with open(temp_file_path, "wb") as f:
        f.write(depth_bytes)
    depth_exr = cv2.imread(temp_file_path, cv2.IMREAD_UNCHANGED)
    os.remove(temp_file_path)

    # change .exr format to .png format
    image_depth = (depth_exr * 1000).astype(np.uint16)
    depth = o3d.geometry.Image(image_depth)

    pcd = image_open3d_to_point_cloud_intrinsic_matrix(
        color, depth, intrinsic_matrix, local_to_world_matrix
    )
    return pcd


def image_array_to_point_cloud_intrinsic_matrix(
    image_rgb: np.ndarray,
    image_depth: np.ndarray,
    intrinsic_matrix: np.ndarray,
    local_to_world_matrix: np.ndarray,
):
    """
    Use the RGB image and depth image, as well as the camera
    intrinsic matrix and extrinsic matrix to generate point cloud in global coordinate.

    Args:
        image_rgb: Numpy.ndarray, in shape (H,W,3), the RGB image.
        image_depth: Numpy.ndarray, in shape (H,W,3), the depth image.
        intrinsic_matrix: Numpy.ndarray, the intrinsic matrix of camera.
        local_to_world_matrix: Numpy.ndarray, the local_to_world_matrix of camera.

    Return:
        open3d.geometry.PointCloud: The point cloud.
    """
    color = o3d.geometry.Image(image_rgb)

    image_depth = (image_depth * 1000).astype(np.uint16)
    depth = o3d.geometry.Image(image_depth)

    pcd = image_open3d_to_point_cloud_intrinsic_matrix(
        color, depth, intrinsic_matrix, local_to_world_matrix
    )

    return pcd


def image_open3d_to_point_cloud_intrinsic_matrix(
    color: o3d.geometry.Image,
    depth: o3d.geometry.Image,
    intrinsic_matrix: np.ndarray,
    local_to_world_matrix: np.ndarray,
):
    """
    Use the RGB image and depth image in open3d.geometry.Image format, as well as the camera
    intrinsic matrix and extrinsic matrix to generate point cloud in global coordinate.

    Args:
        color: open3d.geometry.Image, the RGB image.
        depth: open3d.geometry.Image, the depth image.
        intrinsic_matrix: Numpy.ndarray, the intrinsic matrix of camera.
        local_to_world_matrix: Numpy.ndarray, the local_to_world_matrix of camera.

    Return:
        open3d.geometry.PointCloud: The point cloud.
    """
    rgbd_image = o3d.geometry.RGBDImage.create_from_color_and_depth(
        color, depth, depth_trunc=20, convert_rgb_to_intensity=False
    )

    intrinsic_matrix = o3d.camera.PinholeCameraIntrinsic(
        int(intrinsic_matrix[0, 2] * 2),
        int(intrinsic_matrix[1, 2] * 2),
        intrinsic_matrix[0, 0],
        intrinsic_matrix[1, 1],
        intrinsic_matrix[0, 2],
        intrinsic_matrix[1, 2],
    )

    pcd = o3d.geometry.PointCloud.create_from_rgbd_image(
        rgbd_image, intrinsic_matrix, project_valid_depth_only=False
    )

    # convter to unity space
    pcd.transform([[1, 0, 0, 0], [0, -1, 0, 0], [0, 0, 1, 0], [0, 0, 0, 1]])

    # camera to world in unity space
    pcd.transform(local_to_world_matrix)

    return pcd


def mask_point_cloud_with_id_color(
    pcd: o3d.geometry.PointCloud, image_mask: np.ndarray, color: list
):
    """
    Mask the point cloud with given segmentation masks and target color.

    Args:
        pcd: open3d.geometry.PointCloud, the point cloud.
        image_mask: numpy.ndarray, the segmentation mask in shape (H,W,3).
        color: List, the target color list.

    Return:
        open3d.geometry.PointCloud: The point cloud.
    """
    image_mask = image_mask.reshape(-1, 3)
    index = np.argwhere(image_mask == color)[:, 0]
    index = index[::3]
    id_pcd = o3d.geometry.PointCloud()
    d = np.array(pcd.points)[index]
    id_pcd.points = o3d.utility.Vector3dVector(d)
    c = np.array(pcd.colors)[index]
    id_pcd.colors = o3d.utility.Vector3dVector(c)
    return id_pcd


def mask_point_cloud_with_id_gray_color(
    pcd: o3d.geometry.PointCloud, image_mask: np.ndarray, color: int
):
    """
    Mask the point cloud with given gray-scale segmentation masks and target color.

    Args:
        pcd: open3d.geometry.PointCloud, the point cloud.
        image_mask: numpy.ndarray, the segmentation mask in shape (H,W).
        color: Int, the target gray-scale color.

    Return:
        open3d.geometry.PointCloud: The point cloud.
    """
    image_mask = image_mask.reshape(-1)
    index = np.argwhere(image_mask == color).reshape(-1)
    id_pcd = o3d.geometry.PointCloud()
    d = np.array(pcd.points)[index]
    id_pcd.points = o3d.utility.Vector3dVector(d)
    c = np.array(pcd.colors)[index]
    id_pcd.colors = o3d.utility.Vector3dVector(c)
    return id_pcd


def filter_active_depth_point_cloud_with_exact_depth_point_cloud(
    active_pcd: o3d.geometry.PointCloud,
    exact_pcd: o3d.geometry.PointCloud,
    max_distance: float = 0.05,
):
    """
    Use exact point cloud to filter IR-based active point cloud based on a tolerance distance.

    Args:
        active_pcd: open3d.geometry.PointCloud, the IR-based active point cloud.
        exact_pcd: open3d.geometry.PointCloud, the exact point cloud.
        max_distance: float, the maximum tolerance distance.

    Return:
        open3d.geometry.PointCloud: The point cloud.
    """
    active_point = np.array(active_pcd.points)
    exact_point = np.array(exact_pcd.points)
    # m = exact_point.shape[0]
    # n = active_point.shape[0]
    # duplicated_active_point = np.repeat(active_point, m).reshape((n, m, 3))
    # distance = np.linalg.norm(duplicated_active_point - exact_point, axis=-1)
    distance = np.linalg.norm(active_point - exact_point, axis=-1)
    # min_distance = np.min(distance, axis=1)
    index = np.argwhere(distance < max_distance).reshape(-1)
    filter_pcd = o3d.geometry.PointCloud()
    d = np.array(active_pcd.points)[index]
    filter_pcd.points = o3d.utility.Vector3dVector(d)
    c = np.array(active_pcd.colors)[index]
    filter_pcd.colors = o3d.utility.Vector3dVector(c)
    return filter_pcd



from pyrfuniverse.envs.robotics import FrankaRoboticsEnv


class FrankaReachEnv(FrankaRoboticsEnv):
    def __init__(
        self,
        max_episode_length,
        reward_type,
        executable_file=None,
        scene_file=None,
        asset_bundle_file=None,
        seed=None,
        assets: list = [],
    ):
        super().__init__(
            max_episode_length=max_episode_length,
            reward_type=reward_type,
            tolerance=0.05,
            load_object=False,
            target_in_air=True,
            block_gripper=True,
            target_xz_range=0.15,
            target_y_range=0.6,
            object_xz_range=0.15,
            seed=seed,
            executable_file=executable_file,
            scene_file=scene_file,
            asset_bundle_file=None,
            assets=assets,
        )



from pyrfuniverse.envs import RFUniverseBaseEnv
from pyrfuniverse.envs import RFUniverseGymWrapper
import numpy as np
from gym import spaces
from gym.utils import seeding
import math
import copy


class Robotiq85InsertEnv(RFUniverseGymWrapper):
    def __init__(
        self,
        max_episode_length=40,
        prismatic_factor=0.05,
        insert_factor=0.05,
        hole_range=0.15,
        min_success_reward=0.05,
        executable_file=None,
    ):
        super().__init__(
            executable_file=executable_file,
            articulation_channel=True,
            game_object_channel=True,
        )
        self.max_episode_length = max_episode_length
        self.hole_range = hole_range
        self.min_success_reward = min_success_reward
        self.prismatic_factor = prismatic_factor
        self.insert_factor = insert_factor
        self.bit_wise_factor = np.array(
            [self.prismatic_factor, self.prismatic_factor, self.insert_factor]
        )

        self.t = 0
        self.action_space = spaces.Box(low=-1, high=1, shape=(3,), dtype=np.float32)
        obs = self._get_obs()
        self.observation_space = spaces.Box(
            low=-np.inf, high=np.inf, shape=obs.shape, dtype=np.float32
        )

    def step(self, a: np.ndarray):
        action = a.copy()
        action_ctrl = action * self.bit_wise_factor
        curr_state = self._get_gripper_extra_param()
        target_state = action_ctrl + curr_state

        self._set_target_state(target_state)
        self.t += 1

        obs = self._get_obs()
        reward = self._compute_reward(obs)
        info = {"is_success": reward > self.min_success_reward}
        done = False
        if self.t == self.max_episode_length:
            done = True
            obs = self.reset()

        return obs, reward, done, info

    def reset(self):
        self.env.reset()
        self.t = 0

        hole_x = self._generate_random_float(-self.hole_range, self.hole_range)
        hole_z = self._generate_random_float(-self.hole_range, self.hole_range)
        self.game_object_channel.set_action(
            "SetTransform", index=0, position=[hole_x, 0, hole_z]
        )
        self._step()

        return self._get_obs()

    def seed(self, seed=None):
        self.np_random, seed = seeding.np_random(seed)
        return [seed]

    def render(self, mode="human"):
        self._step()

    def _get_obs(self):
        bar_position = np.array(self.articulation_channel.data[0]["positions"][10])
        bar_velocity = np.array(self.articulation_channel.data[0]["velocities"][10])
        gripper_extra_param = self._get_gripper_extra_param()

        hole_pos = np.array(self.game_object_channel.data[0]["position"])

        obs = np.concatenate(
            (bar_position, bar_velocity, gripper_extra_param, hole_pos)
        )

        return obs

    def _get_gripper_extra_param(self):
        joint_positions = np.array(self.articulation_channel.data[0]["joint_positions"])
        x_prismatic_joint_position = joint_positions[0]
        z_prismatic_joint_position = joint_positions[1]
        bar_prismatic_joint_position = joint_positions[2]

        return np.array(joint_positions)

    def _set_target_state(self, target_state):
        self.articulation_channel.set_action(
            "SetJointPosition", index=0, joint_positions=list(target_state)
        )
        self._step()

    def _compute_reward(self, obs):
        bar_height = obs[1]
        return bar_height * -1

    def _generate_random_float(self, min: float, max: float) -> float:
        assert min <= max, "Min value is {}, while max value is {}.".format(min, max)
        random_float = np.random.rand()
        random_float = random_float * (max - min) + min

        return random_float



from pyrfuniverse.envs.robotics.franka_robotics_env import FrankaRoboticsEnv
from pyrfuniverse.envs.robotics.pick_and_place_env import FrankaPickAndPlaceEnv
from pyrfuniverse.envs.robotics.push_env import FrankaPushEnv
from pyrfuniverse.envs.robotics.reach_env import FrankaReachEnv
from pyrfuniverse.envs.robotics.franka_cloth_env import FrankaClothEnv
from pyrfuniverse.envs.robotics.franka_softbody_env import FrankaSoftbodyEnv
from pyrfuniverse.envs.robotics.franka_cloth_fold_env import FrankaClothFoldEnv

__all__ = [
    "FrankaRoboticsEnv",
    "FrankaReachEnv",
    "FrankaPushEnv",
    "FrankaPickAndPlaceEnv",
    "FrankaClothEnv",
    "FrankaSoftbodyEnv",
    "FrankaClothFoldEnv",
]



import pyrfuniverse.attributes as attr


class GameObjectAttr(attr.BaseAttr):
    """
    Basic game object attribute class.
    """

    def parse_message(self, data: dict):
        """
        Parse messages. This function is called by internal function.

        Returns:
            Dict: A dict containing useful information of this class.

            self.data['3d_bounding_box']: The 3d bounding box of objects.
        """
        super().parse_message(data)

    def SetColor(self, color: list):
        """
        Set object color.

        Args:
            color: A list of length 4, represenging r, g, b and a. Each float is in range (0, 1).
        """
        if color is not None:
            assert len(color) == 4, "color length must be 4"
            color = [float(i) for i in color]

        self._send_data("SetColor", color)

    def EnabledRender(self, enabled: bool):
        """
        Enable or disable rendering system.

        Args:
            enabled: Bool, Ture for enable rendering and False for disable rendering.
        """
        self._send_data("EnabledRender", enabled)

    def SetTexture(self, path: str):
        """
        Set the texture of object.

        Args:
            path: Str, the absolute path for texture file.
        """
        self._send_data("SetTexture", path)

    def Get3DBBox(self):
        """
        Get the 3d bounding box of this object.

        """
        self._send_data("Get3DBBox")



from pyrfuniverse.envs.gym_goal_wrapper_env import RFUniverseGymGoalWrapper
import numpy as np
import math
import copy
from gym import spaces
from gym.utils import seeding


class FrankaClothEnv(RFUniverseGymGoalWrapper):
    metadata = {"render.modes": ["human"]}

    def __init__(
        self,
        asset_bundle_file,
        reward_type,
        tolerance=0.08,
        target_x_range=(-0.1, -0.1),
        target_z_range=(-0.1, 0.1),
        executable_file=None,
    ):
        super().__init__(
            executable_file=executable_file,
            articulation_channel=True,
            game_object_channel=True,
            obi_cloth_channel=True,
        )
        self.asset_bundle_file = asset_bundle_file
        self.reward_type = reward_type
        self.tolerance = tolerance
        self.target_x_range = target_x_range
        self.target_z_range = target_z_range

        # Fixed parameters
        self.obi_solver_position = np.array([-0.1, 0.31, -0.1])
        self.number_of_particles = 289
        self.number_of_sampled_particles = 10

        # Env setup
        self.load_cloth = False
        self._load_obi_cloth()
        self.ik_controller = RFUniverseController(
            "franka", base_pos=np.array([-0.6, 0, 0])
        )
        self.eef_orn = self.ik_controller.bullet_client.getQuaternionFromEuler(
            [math.pi / 2, 0, math.pi / 2]
        )

        self.t = 0
        self.goal = self._sample_goal()
        self.action_space = spaces.Box(low=-1, high=1, shape=(4,), dtype=np.float32)
        obs = self._get_obs()
        self.observation_space = spaces.Dict(
            {
                "observation": spaces.Box(
                    -np.inf, np.inf, shape=obs["observation"].shape, dtype=np.float32
                ),
                "desired_goal": spaces.Box(
                    -np.inf, np.inf, shape=obs["desired_goal"].shape, dtype=np.float32
                ),
                "achieved_goal": spaces.Box(
                    -np.inf, np.inf, shape=obs["achieved_goal"].shape, dtype=np.float32
                ),
            }
        )

    def step(self, action: np.ndarray):
        pos_ctrl = action[:3] * 0.05
        curr_pos = np.array(self.articulation_channel.data[1]["positions"][3])
        pos_ctrl = curr_pos + pos_ctrl
        joint_positions = self.ik_controller.calculate_ik(
            pos_ctrl, eef_orn=self.eef_orn
        )

        curr_gripper_width = self._get_gripper_width()
        target_gripper_width = curr_gripper_width + action[3] * 0.2
        target_gripper_width = np.clip(target_gripper_width, 0, 0.08)
        joint_positions.append(target_gripper_width)

        self._set_franka_joints(np.array(joint_positions))
        self.t += 1

        obs = self._get_obs()
        done = False
        info = {"is_success": self._check_success(obs)}
        reward = self.compute_reward(obs["achieved_goal"], obs["desired_goal"], info)

        return obs, reward, done, info

    def reset(self):
        super().reset()

        self._destroy_obi_cloth()
        self.env.reset()
        self._load_obi_cloth()
        self.goal = self._sample_goal()

        self.t = 0
        self.ik_controller.reset()
        joint_positions = self.ik_controller.calculate_ik([0, 0.3, 0.1], self.eef_orn)
        self.articulation_channel.set_action(
            "SetJointPositionDirectly",
            index=0,
            joint_positions=joint_positions,
        )

        self.game_object_channel.set_action(
            "SetTransform", index=0, position=list(self.goal.copy())
        )

        self._step()

        return self._get_obs()

    def seed(self, seed=None):
        self.np_random, seed = seeding.np_random(seed)
        return [seed]

    def render(self, mode="human"):
        self._step()

    def compute_reward(self, achieved_goal, desired_goal, info):
        distance = self._compute_goal_distance(achieved_goal, desired_goal)
        if self.reward_type == "sparse":
            return -(distance > self.tolerance).astype(np.float32)
        else:
            return -distance

    def _get_obs(self):
        gripper_position = np.array(self.articulation_channel.data[1]["positions"][3])
        gripper_velocity = np.array(self.articulation_channel.data[1]["velocities"][3])
        gripper_width = self._get_gripper_width()
        panda_obs = np.concatenate(
            (gripper_position, gripper_velocity, [gripper_width])
        )

        cloth_particles = np.array(self.obi_cloth_channel.data[0]["positions"])
        random_indices = np.random.randint(
            0, self.number_of_particles, size=(self.number_of_sampled_particles,)
        )
        sampled_particles = cloth_particles[random_indices] + self.obi_solver_position
        sampled_particles = np.reshape(
            sampled_particles, newshape=(3 * self.number_of_sampled_particles,)
        )

        obs = np.concatenate((panda_obs, sampled_particles))

        cloth_center = np.average(cloth_particles, axis=0) + self.obi_solver_position

        return {
            "observation": obs.copy(),
            "achieved_goal": cloth_center.copy(),
            "desired_goal": self.goal.copy(),
        }

    def _load_obi_cloth(self):
        if self.load_cloth:
            return
        self.asset_channel.set_action(
            "LoadObiCloth",
            filename=self.asset_bundle_file,
            name="Obi Solver",
            position=list(self.obi_solver_position),
        )
        self._step()
        self.load_cloth = True

    def _destroy_obi_cloth(self):
        if not self.load_cloth:
            return
        self.obi_cloth_channel.set_action("Destroy", index=0)
        self._step()
        self.load_cloth = False

    def _get_gripper_width(self):
        gripper_joint_positions = copy.deepcopy(
            self.articulation_channel.data[1]["joint_positions"]
        )
        return -1 * (gripper_joint_positions[0] + gripper_joint_positions[1])

    def _set_franka_joints(self, a: np.ndarray):
        self.articulation_channel.set_action(
            "SetJointPosition",
            index=0,
            joint_positions=list(a[0:7]),
        )
        self._step()

        a[7] = -1 * a[7] / 2
        self.articulation_channel.set_action(
            "SetJointPosition",
            index=1,
            joint_positions=[a[7], a[7]],
        )
        self._step()

    def _generate_random_float(self, min: float, max: float) -> float:
        assert min <= max, "Min value is {}, while max value is {}.".format(min, max)
        random_float = np.random.rand()
        random_float = random_float * (max - min) + min

        return random_float

    def _compute_goal_distance(self, goal_a, goal_b):
        assert goal_a.shape == goal_b.shape
        return np.linalg.norm(goal_a - goal_b, axis=-1)

    def _check_success(self, obs):
        achieved_goal = obs["achieved_goal"]
        desired_goal = obs["desired_goal"]
        distance = self._compute_goal_distance(achieved_goal, desired_goal)

        return (distance < self.tolerance).astype(np.float32)

    def _sample_goal(self):
        target_x = self._generate_random_float(
            self.target_x_range[0], self.target_x_range[1]
        )
        target_z = self._generate_random_float(
            self.target_z_range[0], self.target_z_range[1]
        )

        return np.array([target_x, 0.3, target_z])



from pyrfuniverse.envs.gripper_nail.nail_card_env import NailCardEnv
from pyrfuniverse.envs.gripper_nail.nail_can_env import NailCanEnv
from pyrfuniverse.envs.gripper_nail.robotiq85_nail_can_env import Robotiq85NailCanEnv
from pyrfuniverse.envs.gripper_nail.robotiq85_nail_card_env import Robotiq85NailCardEnv
from pyrfuniverse.envs.gripper_nail.robotiq85_nail_book_env import Robotiq85NailBookEnv
from pyrfuniverse.envs.gripper_nail.robotiq85_nail_coin_env import Robotiq85NailCoinEnv
from pyrfuniverse.envs.gripper_nail.robotiq85_insert_env import Robotiq85InsertEnv

__all__ = [
    "NailCardEnv",
    "NailCanEnv",
    "Robotiq85NailCanEnv",
    "Robotiq85NailCardEnv",
    "Robotiq85NailBookEnv",
    "Robotiq85NailCoinEnv",
    "Robotiq85InsertEnv",
]



import pyrfuniverse.attributes as attr


class ColliderAttr(attr.GameObjectAttr):
    """
    Collider class for objects who have collider in Unity.
    """

    def parse_message(self, data: dict):
        """
        Parse messages. This function is called by internal function.

        Returns:
            Dict: A dict containing useful information of this class.
        """
        super().parse_message(data)

    def EnabledAllCollider(self, enabled: bool):
        """
        Set the collider enabled or unenabled.

        Args:
            active: Bool, True for enable and False for unenable.
        """
        self._send_data("EnabledAllCollider", enabled)

    def SetPhysicMaterial(
        self,
        bounciness: float,
        dynamicFriction: float,
        staticFriction: float,
        frictionCombine: int,
        bounceCombine: int,
    ):
        """
        Set the collider physical material.

        Args:
            bounciness (float): The coefficient of restitution or "bounciness" of the collider. It determines how much
                kinetic energy is retained after a collision. A value of 0 means no bounce, while a value of 1 means a
                perfect bounce.
            dynamicFriction (float): The coefficient of friction when the collider is in motion relative to another
                collider. It determines how much resistance there is when the collider is sliding against another surface.
            staticFriction (float): The coefficient of friction when the collider is at rest relative to another collider.
                It determines the resistance to initiating motion between the collider and another surface.
            frictionCombine (int): An integer representing how friction values should be combined when multiple colliders
                interact. It can take on values such as:
                - 0: Average
                - 1: Minimum
                - 2: Maximum
                - 3: Multiply
                These values define how friction will be calculated when multiple colliders are in contact.
            bounceCombine (int): An integer representing how bounce values should be combined when multiple colliders
                interact. It can take on values such as:
                - 0: Average
                - 1: Minimum
                - 2: Maximum
                - 3: Multiply
                These values define how bounciness will be calculated when multiple colliders are in contact.
        """
        self._send_data(
            "SetPhysicMaterial",
            float(bounciness),
            float(dynamicFriction),
            float(staticFriction),
            float(frictionCombine),
            float(bounceCombine),
        )

    def SetRFMoveColliderActive(self, active: bool):
        """
        Set the collider active or inactive in RFMove.

        Args:
            active: Bool, True for active and False for inactive.
        """
        self._send_data("SetRFMoveColliderActive", active)

    def GenerateVHACDColider(self):
        """
        Generate convex colliders using VHACD algorithm.
        """
        self._send_data("GenerateVHACDColider")



from pyrfuniverse.envs.multi_physics.kinova_gen2_catching_cloth_env import (
    KinovaGen2CatchingClothEnv,
)
from pyrfuniverse.envs.multi_physics.ur5_water_shooting import UR5WaterShootingEnv
from pyrfuniverse.envs.multi_physics.flexiv_cutting import FlexivCuttingEnv

__all__ = ["KinovaGen2CatchingClothEnv", "UR5WaterShootingEnv", "FlexivCuttingEnv"]

try:
    from gym.envs.registration import register

    register(
        id="KinovaCatchingCloth-v4",
        entry_point="pyrfuniverse.envs.multi_physics:KinovaGen2CatchingClothEnv",
        kwargs={
            "lock_eef_height": True,
            "with_force_zone": True,
            "max_steps": 100,
            # 'executable_file': join_path('KinovaGen2CatchingCloth/RFUniverse.x86_64')
        },
    )

except ImportError:
    print("No gym installed. Please install gym!")
    pass



import multiprocessing as mp
import os
from collections import OrderedDict
from typing import Any, Callable, List, Optional, Sequence, Tuple, Type, Union

import gym
import numpy as np

from stable_baselines3.common.vec_env.base_vec_env import (
    VecEnv,
    VecEnvIndices,
    VecEnvObs,
    VecEnvStepReturn,
)


def _worker(remote: mp.connection.Connection, env_fn, i) -> None:
    # Import here to avoid a circular import
    from stable_baselines3.common.env_util import is_wrapped

    env = env_fn(proc_id=i)
    while True:
        try:
            cmd, data = remote.recv()
            if cmd == "step":
                observation, reward, done, info = env.step(data)
                if done.all():
                    # save final observation where user can get it, then reset
                    # TODO - bug here
                    # for _i, _o in zip([info, observation]):
                    #     _i["terminal_observation"] = _o
                    observation = env.reset()
                remote.send((observation, reward, done, info))
            elif cmd == "seed":
                remote.send(env.seed(data))
            elif cmd == "reset":
                observation = env.reset()
                remote.send(observation)
            elif cmd == "render":
                remote.send(env.render(data))
            elif cmd == "close":
                env.close()
                remote.close()
                break
            elif cmd == "get_spaces":
                remote.send((env.observation_space, env.action_space))
            elif cmd == "env_method":
                method = getattr(env, data[0])
                remote.send(method(*data[1], **data[2]))
            elif cmd == "get_attr":
                remote.send(getattr(env, data))
            elif cmd == "set_attr":
                remote.send(setattr(env, data[0], data[1]))
            elif cmd == "is_wrapped":
                remote.send(is_wrapped(env, data))
            else:
                raise NotImplementedError(f"`{cmd}` is not implemented in the worker")
        except EOFError:
            break


class SubprocVecEnv(VecEnv):
    """
    Creates a multiprocess vectorized wrapper for multiple environments, distributing each environment to its own
    process, allowing significant speed up when the environment is computationally complex.

    For performance reasons, if your environment is not IO bound, the number of environments should not exceed the
    number of logical cores on your CPU.

    .. warning::

        Only 'forkserver' and 'spawn' start methods are thread-safe,
        which is important when TensorFlow sessions or other non thread-safe
        libraries are used in the parent (see issue #217). However, compared to
        'fork' they incur a small start-up cost and have restrictions on
        global variables. With those methods, users must wrap the code in an
        ``if __name__ == "__main__":`` block.
        For more information, see the multiprocessing documentation.

    :param env_fns: Environments to run in subprocesses
    :param start_method: method used to start the subprocesses.
           Must be one of the methods returned by multiprocessing.get_all_start_methods().
           Defaults to 'forkserver' on available platforms, and 'spawn' otherwise.
    """

    def __init__(
        self,
        env_fns: List[Callable[[], gym.Env]],
        start_method: Optional[str] = None,
        n_agents=32,
    ):
        self.waiting = False
        self.closed = False
        self.n_proc = len(env_fns)
        self.num_envs = self.n_proc * n_agents
        # write n_proc to os env
        os.environ["N_PROC"] = str(self.n_proc)

        if start_method is None:
            # Fork is not a thread safe method (see issue #217)
            # but is more user friendly (does not require to wrap the code in
            # a `if __name__ == "__main__":`)
            forkserver_available = "forkserver" in mp.get_all_start_methods()
            start_method = "forkserver" if forkserver_available else "spawn"
        ctx = mp.get_context(start_method)

        self.remotes, self.work_remotes = zip(*[ctx.Pipe() for _ in range(self.n_proc)])
        self.processes = []
        ids = list(range(self.n_proc))
        for work_remote, env_fn, i in zip(self.work_remotes, env_fns, ids):
            args = (work_remote, env_fn, i)
            # daemon=True: if the main process crashes, we should not cause things to hang
            process = ctx.Process(
                target=_worker, args=args, daemon=True
            )  # pytype:disable=attribute-error
            process.start()
            self.processes.append(process)
            work_remote.close()

        self.remotes[0].send(("get_spaces", None))
        observation_space, action_space = self.remotes[0].recv()
        VecEnv.__init__(self, self.num_envs, observation_space, action_space)

    def step_async(self, actions: np.ndarray) -> None:
        actions = actions.reshape(self.n_proc, -1)

        for remote, action in zip(self.remotes, actions):
            remote.send(("step", action))
        self.waiting = True

    def step_wait(self) -> VecEnvStepReturn:
        results = [remote.recv() for remote in self.remotes]
        self.waiting = False
        obs, rews, dones, infos = zip(*results)
        return (
            _flatten_obs(obs, self.observation_space),
            np.concatenate(rews),
            np.concatenate(dones),
            np.concatenate(infos),
        )

    def seed(self, seed: Optional[int] = None) -> List[Union[None, int]]:
        if seed is None:
            seed = np.random.randint(0, 2**32 - 1)
        for idx, remote in enumerate(self.remotes):
            remote.send(("seed", seed + idx))
        return [remote.recv() for remote in self.remotes]

    def reset(self) -> VecEnvObs:
        for remote in self.remotes:
            remote.send(("reset", None))
        obs = [remote.recv() for remote in self.remotes]
        return _flatten_obs(obs, self.observation_space)

    def close(self) -> None:
        if self.closed:
            return
        if self.waiting:
            for remote in self.remotes:
                remote.recv()
        for remote in self.remotes:
            remote.send(("close", None))
        for process in self.processes:
            process.join()
        self.closed = True

    def get_images(self) -> Sequence[np.ndarray]:
        for pipe in self.remotes:
            # gather images from subprocesses
            # `mode` will be taken into account later
            pipe.send(("render", "rgb_array"))
        imgs = [pipe.recv() for pipe in self.remotes]
        return imgs

    def get_attr(self, attr_name: str, indices: VecEnvIndices = None) -> List[Any]:
        """Return attribute from vectorized environment (see base class)."""
        target_remotes = self._get_target_remotes(indices)
        for remote in target_remotes:
            remote.send(("get_attr", attr_name))
        return [remote.recv() for remote in target_remotes]

    def set_attr(
        self, attr_name: str, value: Any, indices: VecEnvIndices = None
    ) -> None:
        """Set attribute inside vectorized environments (see base class)."""
        target_remotes = self._get_target_remotes(indices)
        for remote in target_remotes:
            remote.send(("set_attr", (attr_name, value)))
        for remote in target_remotes:
            remote.recv()

    def env_method(
        self,
        method_name: str,
        *method_args,
        indices: VecEnvIndices = None,
        **method_kwargs,
    ) -> List[Any]:
        """Call instance methods of vectorized environments."""
        target_remotes = self._get_target_remotes(indices)
        for remote in target_remotes:
            remote.send(("env_method", (method_name, method_args, method_kwargs)))
        return [remote.recv() for remote in target_remotes]

    def env_is_wrapped(
        self, wrapper_class: Type[gym.Wrapper], indices: VecEnvIndices = None
    ) -> List[bool]:
        """Check if worker environments are wrapped with a given wrapper"""
        target_remotes = self._get_target_remotes(indices)
        for remote in target_remotes:
            remote.send(("is_wrapped", wrapper_class))
        return [remote.recv() for remote in target_remotes]

    def _get_target_remotes(self, indices: VecEnvIndices) -> List[Any]:
        """
        Get the connection object needed to communicate with the wanted
        envs that are in subprocesses.

        :param indices: refers to indices of envs.
        :return: Connection object to communicate between processes.
        """
        indices = self._get_indices(indices)
        return [self.remotes[i] for i in indices]


def _flatten_obs(
    obs: Union[List[VecEnvObs], Tuple[VecEnvObs]], space: gym.spaces.Space
) -> VecEnvObs:
    """
    Flatten observations, depending on the observation space.

    :param obs: observations.
                A list or tuple of observations, one per environment.
                Each environment observation may be a NumPy array, or a dict or tuple of NumPy arrays.
    :return: flattened observations.
            A flattened NumPy array or an OrderedDict or tuple of flattened numpy arrays.
            Each NumPy array has the environment index as its first axis.
    """
    assert isinstance(
        obs, (list, tuple)
    ), "expected list or tuple of observations per environment"
    assert len(obs) > 0, "need observations from at least one environment"

    if isinstance(space, gym.spaces.Dict):
        assert isinstance(
            space.spaces, OrderedDict
        ), "Dict space must have ordered subspaces"
        assert isinstance(
            obs[0], dict
        ), "non-dict observation for environment with Dict observation space"
        return OrderedDict(
            # vstack ! for multiple agents in the sub env
            [(k, np.vstack([o[k] for o in obs])) for k in space.spaces.keys()]
        )
    elif isinstance(space, gym.spaces.Tuple):
        assert isinstance(
            obs[0], tuple
        ), "non-tuple observation for environment with Tuple observation space"
        obs_len = len(space.spaces)
        return tuple(np.stack([o[i] for o in obs]) for i in range(obs_len))
    else:
        return np.stack(obs)



from setuptools import setup
from setuptools import find_packages

import pyrfuniverse

VERSION = pyrfuniverse.__version__

setup(
    name="pyrfuniverse",
    version=VERSION,
    description="RFUniverse python interface",
    url="https://github.com/mvig-robotflow/pyrfuniverse",
    author="Robotflow AI Team",
    author_email="robotflow@163.com",
    packages=find_packages(exclude=["*.tests", "*.tests.*", "tests.*", "tests"]),
    zip_safe=False,
    install_requires=[
        "numpy>=1.14.1",
        "opencv-contrib-python",
    ],
    python_requires=">=3.6.1",
)



import pyrfuniverse.attributes as attr


class SoftbodyAttr(attr.BaseAttr):
    """
    Obi Softbody class
    """

    def parse_message(self, data: dict):
        """
        Parse messages. This function is called by internal function.

        Returns:
            Dict: A dict containing useful information of this class.
        """
        super().parse_message(data)



from pyrfuniverse.envs import RFUniverseBaseEnv
from pyrfuniverse.envs import RFUniverseGymWrapper
from pyrfuniverse.side_channel import SideChannel, IncomingMessage
import numpy as np
from gym import spaces
from gym.utils import seeding
import uuid


class CollisionDetectionChannel(SideChannel):
    def __init__(self):
        super().__init__(uuid.UUID("bee25cbc-07e2-11ec-9e67-18c04d443e7d"))
        self.num_collision = 0

    def on_message_received(self, msg: IncomingMessage) -> None:
        self.num_collision += msg.read_int32()

    def get_num_collision(self):
        num_collision = self.num_collision
        self.clear()

        return num_collision

    def clear(self):
        self.num_collision = 0


class MultiAgentNavigationEnv(RFUniverseGymWrapper):
    metadata = {"render.modes": ["human"]}

    def __init__(
        self,
        num_agents,
        asset_bundle_file,
        reset_on_collision=False,
        max_episode_length=100,
        executable_file=None,
    ):
        self.collision_detection_channel = CollisionDetectionChannel()
        super().__init__(
            executable_file=executable_file,
            custom_channels=[self.collision_detection_channel],
            rigidbody_channel=True,
        )
        self.num_agents = num_agents
        self.asset_bundle_file = asset_bundle_file
        self.max_episode_length = max_episode_length
        self.reset_on_collision = reset_on_collision

        # Fixed parameters
        self.agent_name = "NavRobot"
        self.y_offset = 0.15
        self.reset_agent_min_distance = 1.1
        self.world_range_low = np.array([-4, self.y_offset, -4])
        self.world_range_high = np.array([4, self.y_offset, 4])
        self.strength = 1000
        self.collision_multiplier = 50
        self.seed()

        self.action_space = spaces.Box(
            low=-1, high=1, shape=(2 * self.num_agents,), dtype=np.float32
        )
        self._env_setup()
        obs = self._get_obs()
        self.observation_space = spaces.Box(
            low=-np.inf, high=np.inf, shape=obs.shape, dtype=np.float32
        )
        self.t = 0

        self.accumulative_collisions = 0

    def step(self, action: np.ndarray):
        a = action.copy()
        # We promise each agent will move at least 0.1m on x-axis and z-axis
        # a = a * 0.4
        # for i in range(action.shape[0]):
        #     if a[i] >= 0:
        #         a[i] += 0.1
        #     else:
        #         a[i] -= 0.1

        for i in range(self.num_agents):
            target_force = np.array([a[2 * i], 0, a[2 * i + 1]]) * self.strength
            self._set_agent_force(i, target_force)

        reward = self._get_total_velocity() / self.num_agents
        num_collisions = self.collision_detection_channel.get_num_collision()
        reward += -1 * num_collisions * self.collision_multiplier
        self.accumulative_collisions += num_collisions

        self.t += 1
        done = False
        info = {"is_success": num_collisions < 1}

        if self.reset_on_collision and num_collisions > 0:
            self.reset()
            done = True

        elif self.t == self.max_episode_length:
            self.reset()
            done = True

        return self._get_obs(), reward, done, info

    def reset(self):
        # print(self.accumulative_collisions)
        self.env.reset()
        self.t = 0
        self.accumulative_collisions = 0

        positions = []
        for i in range(self.num_agents):
            while True:
                agent_position = self.np_random.uniform(
                    self.world_range_low, self.world_range_high
                )
                if self._check_reset_position_legality(positions, agent_position):
                    break
            positions.append(agent_position.copy())
            self._set_agent_position(i, agent_position)

        # Ignore the collision when organizing agents
        self.collision_detection_channel.clear()

        return self._get_obs()

    def seed(self, seed=None):
        self.np_random, seed = seeding.np_random(seed)
        return [seed]

    def render(self, mode="human"):
        self._step()

    def _env_setup(self):
        for _ in range(self.num_agents):
            agent_position = self.np_random.uniform(
                self.world_range_low, self.world_range_high
            )
            self.asset_channel.set_action(
                "LoadRigidbody",
                filename=self.asset_bundle_file,
                name=self.agent_name,
                position=list(agent_position),
            )
            self._step()

    def _get_obs(self):
        agent_obs = np.array([])
        for i in range(self.num_agents):
            agent_position = self.rigidbody_channel.data[i]["position"]
            agent_velocity = self.rigidbody_channel.data[i]["velocity"]
            # We ignore y-axis
            this_agent_obs = np.array(
                [
                    agent_position[0],
                    agent_position[2],
                    agent_velocity[0],
                    agent_velocity[2],
                ]
            )
            agent_obs = np.concatenate((agent_obs, this_agent_obs))

        return agent_obs.copy()

    def _set_agent_position(self, index, position):
        self.rigidbody_channel.set_action(
            "SetTransform", index=index, position=list(position), rotation=[0, 0, 0]
        )
        self._step()

    def _set_agent_force(self, index, force):
        self.rigidbody_channel.set_action("AddForce", index=index, force=force)
        self._step()

    def _get_total_velocity(self):
        total_velocity = 0
        for i in range(self.num_agents):
            velocity = self.rigidbody_channel.data[i]["velocity"]
            total_velocity += np.linalg.norm(
                np.array([velocity[0], velocity[2]]), axis=-1
            )

        return total_velocity

    def _check_reset_position_legality(self, positions, agent_pos):
        for position in positions:
            if np.linalg.norm(position - agent_pos) < self.reset_agent_min_distance:
                return False

        return True



from pyrfuniverse.envs import RFUniverseGymWrapper
from pyrfuniverse.side_channel import SideChannel, IncomingMessage
import numpy as np
from gym import spaces
from gym.utils import seeding
import uuid


class SuccessChannel(SideChannel):
    def __init__(self):
        super().__init__(uuid.UUID("f32e9f8c-b04d-4720-afe6-d7574ea713a5"))
        self.data = {}

    def on_message_received(self, msg: IncomingMessage) -> None:
        self.data["success"] = msg.read_bool()
        self.data["finish_last_step"] = msg.read_bool()


class BouncerEnv(RFUniverseGymWrapper):
    metadata = {"render.modes": ["human"]}

    def __init__(self, executable_file=None):
        self.success_channel = SuccessChannel()
        super().__init__(
            executable_file,
            custom_channels=[self.success_channel],
            rigidbody_channel=True,
        )
        self.action_space = spaces.Box(
            low=np.array([-1, -1, -1]), high=np.array([1, 1, 1]), dtype=np.float32
        )
        self.observation_space = spaces.Box(
            low=-np.inf, high=np.inf, shape=(6,), dtype=np.float32
        )
        self.strength = 500
        self.t = 0
        self.max_steps = 20

    def step(self, a: np.ndarray):
        force = a.copy()
        force[1] = (force[1] + 1) / 2 + 1
        self.env_param_channel.set_float_parameter("lookDirX", force[0])
        self.env_param_channel.set_float_parameter("lookDirY", force[1] - 1)
        self.env_param_channel.set_float_parameter("lookDirZ", force[2])

        force = force * self.strength
        self.rigidbody_channel.set_action("AddForce", index=0, force=list(force))

        self.success_channel.data = {}
        success = False
        done = False
        reward = -0.05 * (np.linalg.norm(a) ** 2) / 3
        info = {}

        while (
            self.success_channel.data == {}
            or self.success_channel.data["finish_last_step"] == False
        ):
            self._step()
            if (
                "success" in self.success_channel.data.keys()
                and self.success_channel.data["success"]
            ):
                success = True
            if self._check_fail():
                success = False
                reward = -1
                done = True
                break

        if success:
            reward = 1
            info["is_success"] = True
        else:
            info["is_success"] = False

        self.t += 1
        if self.t == self.max_steps:
            done = True

        if done:
            self.reset()

        return self._get_obs(), reward, done, info

    def _check_fail(self):
        position = self.rigidbody_channel.data[0]["position"]
        if position[1] < -1:
            return True
        if (
            position[0] < -19
            or position[0] > 19
            or position[2] < -19
            or position[2] > 19
        ):
            return True

        return False

    def _get_obs(self):
        object_pos = self.rigidbody_channel.data[0]["position"]
        object_pos = np.array(object_pos)
        target_pos = self.rigidbody_channel.data[1]["position"]
        target_pos = np.array(target_pos)

        return np.concatenate((object_pos, target_pos))

    def reset(self):
        self.t = 0
        self.env_param_channel.set_float_parameter(
            "bouncerPosX", self._generate_random_float(-5, 5)
        )
        self.env_param_channel.set_float_parameter(
            "bouncerPosZ", self._generate_random_float(-5, 5)
        )
        self.env_param_channel.set_float_parameter(
            "targetPosX", self._generate_random_float(-5, 5)
        )
        self.env_param_channel.set_float_parameter(
            "targetPosY", self._generate_random_float(2, 7)
        )
        self.env_param_channel.set_float_parameter(
            "targetPosZ", self._generate_random_float(-5, 5)
        )
        self.env_param_channel.set_float_parameter("lookDirX", 0)
        self.env_param_channel.set_float_parameter("lookDirY", 0)
        self.env_param_channel.set_float_parameter("lookDirZ", 0)

        self.env.reset()

        return self._get_obs()

    def seed(self, seed=1234):
        self.np_random, seed = seeding.np_random(seed)
        return [seed]

    def render(self, mode="human"):
        self._step()

    def _generate_random_float(self, min: float, max: float) -> float:
        assert min < max, "Min value is {}, while max value is {}.".format(min, max)
        random_float = np.random.rand()
        random_float = random_float * (max - min) + min

        return random_float


class BouncerEnvV0(BouncerEnv):
    def __init__(self):
        super().__init__(
            "/home/haoyuan/workspace/rfuniverse/build/Bouncer/RFUniverse.x86_64"
        )



import pyrfuniverse.attributes as attr


class GraspSimAttr(attr.BaseAttr):
    """
    Grasp pose simulation class.
    """

    def parse_message(self, data: dict):
        """
        Parse messages. This function is called by internal function.

        Returns:
            Dict: A dict containing useful information of this class.

            self.data['done']: Whether the simulation is done

            self.data['points']: The list of grasp points.

            self.data['quaternions']: The list of grasping pose quaternions.

            self.data['width']: The list of gripper width of grasping pose.

            self.data['success']: The list of success or failure of the grasing pose.
        """
        super().parse_message(data)

    def StartGraspSim(
        self,
        mesh: str,
        gripper: str,
        points: list,
        normals: list,
        depth_range_min: float,
        depth_range_max: float,
        depth_lerp_count: int,
        angle_lerp_count: int,
        parallel_count: int = 100,
    ):
        """
        Start simulating grasping.

        Args:
            mesh: Str, the absolute path to .obj file.
            gripper: Str, the name of the gripper.
            points: A list of float, representing the grasping points.
            normals: A list of float, representing the normals.
            depth_range_min: Float, the minimum depth of grasp pose.
            depth_range_max: Float, the maximum depth of grasp pose.
            depth_lerp_count: Int, the interpolation count of depth.
            angle_lerp_count: Int, the interpolation count of angle.
            parallel_count: Int, the count of parallel grasping.
        """
        self._send_data(
            "StartGraspSim",
            mesh,
            gripper,
            points,
            normals,
            float(depth_range_min),
            float(depth_range_max),
            depth_lerp_count,
            angle_lerp_count,
            parallel_count,
        )

    def GenerateGraspPose(
        self,
        mesh: str,
        gripper: str,
        points: list,
        normals: list,
        depth_range_min: float,
        depth_range_max: float,
        depth_lerp_count: int,
        angle_lerp_count: int,
    ):
        """
        Generate grasp poses and visualize grasp results.

        Args:
            mesh: Str, the absolute path to .obj file.
            gripper: Str, the name of the gripper.
            points: A list of float, representing the grasping points.
            normals: A list of float, representing the normals.
            depth_range_min: Float, the minimum depth of grasp pose.
            depth_range_max: Float, the maximum depth of grasp pose.
            depth_lerp_count: Int, the interpolation count of depth.
            angle_lerp_count: Int, the interpolation count of angle.
        """
        self._send_data(
            "GenerateGraspPose",
            mesh,
            gripper,
            points,
            normals,
            float(depth_range_min),
            float(depth_range_max),
            depth_lerp_count,
            angle_lerp_count,
        )

    def StartGraspTest(
        self,
        mesh: str,
        gripper: str,
        points: list,
        quaternions: list,
        parallel_count: int = 100,
    ):
        """
        Start testing the grasp based on current grasp poses.

        Args:
            mesh: Str, the absolute path to .obj file.
            gripper: Str, the name of the gripper.
            points: A list of float, representing the grasping points.
            quaternions: A list of float, representing the quaternions.
            parallel_count: Int, the interpolation count of angle.
        """
        self._send_data(
            "StartGraspTest", mesh, gripper, points, quaternions, parallel_count
        )

    def ShowGraspPose(
        self, mesh: str, gripper: str, positions: list, quaternions: list
    ):
        """
        Display grasp poses.

        Args:
            mesh: Str, the absolute path to .obj file.
            gripper: Str, the name of the gripper.
            points: A list of float, representing the grasping points.
            quaternions: A list of float, representing the quaternions.
        """
        self._send_data("ShowGraspPose", mesh, gripper, positions, quaternions)



class BaseAttr:
    """
    Base attribute class, which includes general functions such as
    object loading, deleting and transforming.
    """

    def __init__(self, env, id: int, data=None):
        if data is None:
            data = {}
        self.env = env
        self.id = id
        self.data = data

    def parse_message(self, data: dict):
        """
        Parse messages. This function is called by internal function.

        Returns:
            Dict: A dict containing useful information of this class.

            self.data['name']: The name of the object.

            self.data['position']: The position of the object in world coordinate.

            self.data['rotation']: The euler angle of the object in world coordinate.

            self.data['quaternion']: The quaternion of the object in world coordinate.

            self.data['local_position']: The position of the object in its parent's local coordinate.

            self.data['local_rotation']: The euler angle of the object in its parent's local coordinate.

            self.data['local_quaternion']: The quaternion of the object in its parent's local coordinate.

            self.data['local_to_world_matrix']: The transformation matrix from local to world coordinate.

            self.data['result_local_point']: The result of transforming object from local to world coordinate.

            self.data['result_world_point']: The result of transforming object from world to local coordinate.
        """
        self.data = data
        return self.data

    def _send_data(self, message: str, *args):
        self.env._send_instance_data(self.id, message, *args)

    def SetType(self, attr_type: type):
        """
        Set the attribute type of this object

        Args:
            attr_type: Any attribute in pyrfuniverse.attributes.

        Returns:
            The target attribute.
        """
        self.env.attrs[self.id] = attr_type(self.env, self.id, self.data)
        return self.env.attrs[self.id]

    def SetTransform(
        self,
        position: list = None,
        rotation: list = None,
        scale: list = None,
        is_world: bool = True,
    ):
        """
        Set the transform of this object, including position, rotation, scale and coordinate.

        Args:
            position: A list of length 3, representing the target position value of object.
            rotation: A list of length 3, representing the target euler angle value of object.
            scale: A list of length 3, representing the target scale value of object.
            is_world: Bool, True for world coordinate, False for local coordinate.
        """
        if position is not None:
            assert len(position) == 3, "position length must be 3"
            position = [float(i) for i in position]
        if rotation is not None:
            assert len(rotation) == 3, "rotation length must be 3"
            rotation = [float(i) for i in rotation]
        if scale is not None:
            assert len(scale) == 3, "scale length must be 3"
            scale = [float(i) for i in scale]

        self._send_data("SetTransform", position, rotation, scale, is_world)

    def SetPosition(self, position: list = None, is_world: bool = True):
        """
        Set the position of this object

        Args:
            position: A list of length 3, representing the target position value of object.
            is_world: Bool, True for world coordinate, False for local coordinate.
        """
        assert len(position) == 3, "position length must be 3"
        position = [float(i) for i in position]

        self._send_data("SetPosition", position, is_world)

    def SetRotation(self, rotation: list = None, is_world: bool = True):
        """
        Set the rotation of this object

        Args:
            rotation: A list of length 3, representing the target euler angle value of object.
            is_world: Bool, True for world coordinate, False for local coordinate.
        """
        assert len(rotation) == 3, "rotation length must be 3"
        rotation = [float(i) for i in rotation]

        self._send_data("SetRotation", rotation, is_world)

    def SetRotationQuaternion(self, quaternion: list = None, is_world: bool = True):
        """
        Rotate this object using quaternion.

        Args:
            quaternion: A list of length 4, representing the quaternion from current pose.
            is_world: Bool, True for world coordinate, False for local coordinate.
        """
        assert len(quaternion) == 4, "quaternion length must be 4"
        quaternion = [float(i) for i in quaternion]

        self._send_data("SetRotationQuaternion", quaternion, is_world)

    def SetScale(self, scale: list = None):
        """
        Set the scale of this object

        Args:
            scale: A list of length 3, representing the target scale value of object.
        """
        assert len(scale) == 3, "scale length must be 3"
        scale = [float(i) for i in scale]

        self._send_data("SetScale", scale)

    def Translate(self, translation: list, is_world: bool = True):
        """
        Translate this object.

        Args:
            translation: A list of length 3, representing the translation from current position.
            is_world: Bool, True for world coordinate, False for local coordinate.
        """
        assert len(translation) == 3, "translation length must be 3"
        translation = [float(i) for i in translation]

        self._send_data("Translate", translation, is_world)

    def Rotate(self, rotation: list, is_world: bool = True):
        """
        Rotate this object.

        Args:
            rotation: A list of length 3, representing the euler-angle-format rotation from current euler angle.
            is_world: Bool, True for world coordinate, False for local coordinate.
        """
        assert len(rotation) == 3, "rotation length must be 3"
        rotation = [float(i) for i in rotation]

        self._send_data("Rotate", rotation, is_world)

    def LookAt(self, target: list, world_up: list = None):
        """
        Rotates the transform so the forward vector points at target's current position.

        Args:
            target: A list of length 3, target to point towards.
            world_up: A list of length 3, vector specifying the upward direction.
        """
        if world_up is None:
            world_up = [0.0, 1.0, 0.0]
        assert len(target) == 3, "target length must be 3"
        target = [float(i) for i in target]
        assert len(world_up) == 3, "world_up length must be 3"
        world_up = [float(i) for i in world_up]

        self._send_data("LookAt", target, world_up)

    def SetActive(self, active: bool):
        """
        Set the activeness of this obeject.

        Args:
            active: Bool, True for active, False for inactive.
        """
        self._send_data("SetActive", active)

    def SetParent(self, parent_id: int, parent_name: str = ""):
        """
        Set the parent of this object.

        Args:
            parent_id: Int, the id of parent object.
            parent_name: Str, the name of parent object.
        """
        self._send_data("SetParent", parent_id, parent_name)

    def SetLayer(self, layer: int):
        """
        Set the layer in Unity of this object.

        Args:
            layer: Int, the number of layer.
        """
        self._send_data("SetLayer", layer)

    def Copy(self, new_id: int):
        """
        Duplicate an object.

        Args:
            new_id: Int, the id of new object.
        """
        self._send_data("Copy", new_id)

        self.env.attrs[new_id] = type(self)(self.env, new_id, self.data)
        return self.env.attrs[new_id]

    def Destroy(self):
        """
        Destroy this object in Unity.
        """
        self._send_data("Destroy")

        self.env.attrs.pop(self.id)

    def GetLocalPointFromWorld(self, point: list):
        """
        Transform a point from local coordinate to world coordinate. After calling this method and stepping once, the result will be saved in self.data['result_local_point']

        Args:
            point: A list of length 3, representing the position of a point.
        """
        assert len(point) == 3, "point length must be 3"
        point = [float(i) for i in point]

        self._send_data("GetLocalPointFromWorld", point)

    def GetWorldPointFromLocal(self, point: list):
        """
        Transform a point from world coordinate to local coordinate. After calling this method and stepping once, the result will be saved in self.data['result_world_point']

        Args:
            point: A list of length 3, representing the position of a point.
        """
        assert len(point) == 3, "point length must be 3"
        point = [float(i) for i in point]

        self._send_data("GetWorldPointFromLocal", point)
